library ieee;
use ieee.std_logic_1164.all;

use ieee.numeric_std.all;
use std.textio.all;
use work.all;

entity ufunc is 
port (NxDI :   in std_logic_vector(0 to 255); 
      PxDO:   out std_logic_vector(0 to 255) 
     );
end entity ufunc;

architecture uf of ufunc is

subtype Int6Type is integer range 0 to 200;

constant mul : Int6Type :=128;
constant r1 : Int6Type := 80;
constant r2 : Int6Type := 88;
constant r3 : Int6Type := 88;

constant al : Int6Type := 68;
constant bl : Int6Type := 64;
constant cl : Int6Type := 68;

constant af : Int6Type := 74;
constant bf : Int6Type := 66;
constant cf : Int6Type := 84;

constant aa1 : Int6Type := 73;
constant aa2 : Int6Type := 79;

constant ba1 : Int6Type := 65;
constant ba2 : Int6Type := 87;

constant ca1 : Int6Type := 77;
constant ca2 : Int6Type := 87;

constant be  : Int6Type := 85;
constant ce  : Int6Type := 85;

signal R1xD : std_logic_vector (0 to r1+mul-1);
signal R2xD : std_logic_vector (0 to r2+mul-1);
signal R3xD : std_logic_vector (0 to r3+mul-1);

signal A1xD,A2xD,A3xD : std_logic_vector (0 to mul-1);

 


begin

R1xD(mul to r1+mul-1) <= NxDI(0 to r1-1);

R2xD(mul to r2+mul-1) <= NxDI(r1 to r1+r2-1);

R3xD(mul to r3+mul-1) <= NxDI(r1+r2 to r1+r2+r3-1);

 

a1: for i in 0 to mul-1 generate 

A1xD(i) <= R1xD((al-1) + mul - i) xor  R1xD((r1-1) + mul -i)  xor (R3xD((ce+r1+r2-1) + (mul-r1-r2) -i ) and R2xD((be+r1-1) + (mul-r1) -i));  
A2xD(i) <= R2xD((bl+r1-1) + (mul-r1) -i)   xor  R2xD((r1+r2-1) + (mul-r1) -i);
A3xD(i) <= R3xD((cl+r1+r2-1) + (mul-r1-r2) -i)  xor  R3xD((r1+r2+r3-1) + (mul-r1-r2) -i)  ; 

R2xD(mul - i-1) <= (R1xD((aa1-1)+  mul -i  )       and R1xD((aa2-1)+ mul -i )        )       xor  A1xD(i)  xor   R2xD((bf+r1-1) + (mul-r1) -i )  ;
R3xD(mul - i-1) <= (R2xD((ba1+r1-1) + (mul-r1) -i )  and R2xD((ba2+r1-1) + (mul-r1) -i ) )       xor  A2xD(i)  xor   R3xD((cf+r1+r2-1) + (mul-r1-r2) -i )  ;
R1xD(mul - i-1) <= (R3xD((ca1+r1+r2-1) + (mul-r1-r2) -i ) and R3xD((ca2+r1+r2-1) + (mul-r1-r2) -i) )   xor  A3xD(i)  xor   R1xD(af-1  + mul - i  )  ;

 
end generate a1;

PxDO <= R1xD(0 to r1-1) & R2xD(0 to r2-1) & R3xD(0 to r3-1);

end architecture uf;


