------------------------------------------------------------------------------- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use std.textio.all;
use ieee.std_logic_textio.all;
--use work.aes_pack.all;

entity triad_tb is
end triad_tb;


architecture tb of triad_tb is   

signal KeyxD :   std_logic_vector(0 to 127); 
signal IVxD:     std_logic_vector(0 to 95);
signal INxD :    std_logic;
signal CSxD :    std_logic_vector(0 to 1);
 
signal ClkxC:    std_logic; 
signal ResxRB:   std_logic; 
signal CTxD:     std_logic;  
signal TagxD:    std_logic ;
signal ECxS:     std_logic;   
signal ETxS:     std_logic;

signal    LenPTxD:        std_logic_vector (7 downto 0);

signal    FAxD:           std_logic_vector (1023 downto 0); 
signal    FPxD:           std_logic_vector (32767 downto 0);

signal    LenADxD:        std_logic_vector (2 downto 0);

  component Triad
  port(
       KeyxDI : in std_logic_vector(0 to 127); 
       IVxDI:   in std_logic_vector(0 to 95);
       INxDI:   in std_logic;
       CSxDI:   in std_logic_vector(0 to 1);
      
       ClkxCI:  in std_logic; 
       ResxRBI: in std_logic; 
       CTxDO:   out std_logic;  
       TagxDO:  out std_logic ;
       ECxSO:   out std_logic;   
       ETxSO:   out std_logic
  );
  end component;
       


   
begin

  -- Instantiate the module under test (MUT)
  mut: Triad
    port map (
        KeyxDI => KeyxD,
        IVxDI  => IVxD,
	INxDI=> INxD,
	CSxDI=> CSxD,
        ClkxCI => ClkxC,
        ResxRBI=> ResxRB,
	CTxDO=> CTxD ,
	TagxDO=> TagxD,
	ECxSO=> ECxS ,
	ETxSO=> ETxS 
 
  );

  -- Generate the clock
--  ClkxC <= not (ClkxC) after clkphasehigh;


  Tb_clkgen : process
  begin
     ClkxC <= '1';
     wait for 50 ns;
     ClkxC <= '0';
     wait for 50 ns;
  end process Tb_clkgen;

  -- obtain stimulus and apply it to MUT
  ----------------------------------------------------------------------------
  B1 : block
   begin


  -- timing of clock and simulation events
 


  Tb_stimappli : process
    variable INLine   : line;
    variable temp_k   : STD_LOGIC_VECTOR(0 to 127);
    variable temp_p   : STD_LOGIC_VECTOR(0 to 95); 
    variable temp_t   : STD_LOGIC_VECTOR(32767 downto 0);
    variable temp_a   : STD_LOGIC_VECTOR(32767 downto 0); 
    variable temp_len : integer range 0 to 4096;  
    variable tempa_len : integer range 0 to 128;
    variable int_len,blen :  integer range 0 to 32767;
    variable inta_len,balen :  integer range 0 to 8;
    variable rem_len,rema_len : integer range 0 to 15;
 
    constant clkphasehigh: time:= 50 ns;
    constant clkphaselow: time:= 50 ns;
     constant clkper: time:= 100 ns;
    constant resetactivetime:         time:= 25 ns;

 
 

  file INPUT_file,keyfile,ivfile,adfile,ptfile,ladfile,lptfile : TEXT;



  begin
    -- Opening Input File
    file_open(keyfile, "Key.txt", read_mode);
    file_open(ivfile,  "IV.txt", read_mode);
    file_open(adfile,  "IN1.txt", read_mode);
    file_open(ptfile,  "CS1.txt", read_mode);
 
    -- default values

    -- Reset phase
    -- Reset is active 
    ResxRB      <= '0';
    IVxD <= (others => '0');
    KeyxD <= (others => '0');
    INxD <= '0';
    CSxD <= (others => '0');
    wait for resetactivetime;
    ResxRB      <= '1';

    -- process until we run out of stimuli

    appli_loop : while not (endfile(keyfile)) loop
   
      readline(keyfile,INLine);
      hread(INLine,temp_k);
      KeyxD <= temp_k;
      readline(ivfile,INLine);
      hread(INLine,temp_p);
      IVxD <= temp_p;

      wait for clkper-resetactivetime;

    
    readline(adfile,INLine);
    blen := INLine'length;
     
    hread(INLine,temp_a(4*blen-1 downto 0)) ;
    readline(ptfile,INLine);
    --hread(INLine,temp_p(4*blen-1 downto 0)) ; 
    
    for lv in 0 to blen-1 loop
       
        if INLine(lv+1)= '0' then 
           INxD<= temp_a(4*blen-1 -4*lv);
           CSxD<= "00";
           wait for clkper;
	   INxD<= temp_a(4*blen-1 -4*lv-1);
           CSxD<= "00";
           wait for clkper;
  	   INxD<= temp_a(4*blen-1 -4*lv-2);
           CSxD<= "00";
           wait for clkper;
           INxD<= temp_a(4*blen-1 -4*lv-3);
           CSxD<= "00";
           wait for clkper;
        elsif INLine(lv+1)= '1' then 
           INxD<= temp_a(4*blen-1 -4*lv);
           CSxD<= "01";
           wait for clkper;
	   INxD<= temp_a(4*blen-1 -4*lv-1);
           CSxD<= "01";
           wait for clkper;
  	   INxD<= temp_a(4*blen-1 -4*lv-2);
           CSxD<= "01";
           wait for clkper;
           INxD<= temp_a(4*blen-1 -4*lv-3);
           CSxD<= "01";
           wait for clkper;
        elsif INLine(lv+1)= '2' then 
           INxD<= temp_a(4*blen-1 -4*lv);
           CSxD<= "01";
           wait for clkper;
	   INxD<= temp_a(4*blen-1 -4*lv-1);
           CSxD<= "01";
           wait for clkper;
  	   INxD<= temp_a(4*blen-1 -4*lv-2);
           CSxD<= "01";
           wait for clkper;
           INxD<= temp_a(4*blen-1 -4*lv-3);
           CSxD<= "10";
           wait for clkper;
        end if;
    end loop;
    ResxRB      <= '0';
    wait for resetactivetime;
    ResxRB      <= '1';

    end loop appli_loop;
    wait until ClkxC'event and ClkxC = '1';
    file_close(INPUT_file);
    wait;
  end process Tb_stimappli;
end block;
end tb;

